package net.minecraftforge.fluids;

import net.minecraft.block.Block;
import net.minecraft.block.material.Material;
import net.minecraft.block.state.IBlockState;
import net.minecraft.creativetab.CreativeTabs;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.init.Blocks;
import net.minecraft.init.SoundEvents;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.stats.StatList;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.RayTraceResult;
import net.minecraft.util.text.translation.I18n;
import net.minecraft.util.ActionResult;
import net.minecraft.util.EnumActionResult;
import net.minecraft.util.EnumHand;
import net.minecraft.util.EnumParticleTypes;
import net.minecraft.util.SoundCategory;
import net.minecraft.world.World;
import net.minecraftforge.event.entity.player.FillBucketEvent;
import net.minecraftforge.fml.common.eventhandler.Event;
import net.minecraftforge.fml.common.eventhandler.EventPriority;
import net.minecraftforge.fml.common.eventhandler.SubscribeEvent;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;
import net.minecraftforge.items.ItemHandlerHelper;

import java.util.List;

/**
 * A universal bucket that can hold any liquid
 */
public class UniversalBucket extends Item implements IFluidContainerItem
{

    public final int capacity; // how much the bucket holds
    public final ItemStack empty; // empty item to return and recognize when filling
    public final boolean nbtSensitive;

    public UniversalBucket()
    {
        this(FluidContainerRegistry.BUCKET_VOLUME, FluidContainerRegistry.EMPTY_BUCKET, false);
    }

    /**
     * @param capacity        Capacity of the container
     * @param empty           Item used for filling with the bucket event and returned when emptied
     * @param nbtSensitive    Whether the empty item is NBT sensitive (usually true if empty and full are the same items)
     */
    public UniversalBucket(int capacity, ItemStack empty, boolean nbtSensitive)
    {
        this.capacity = capacity;
        this.empty = empty;
        this.nbtSensitive = nbtSensitive;

        this.func_77625_d(1);

        this.func_77637_a(CreativeTabs.field_78026_f);
    }

    @SideOnly(Side.CLIENT)
    @Override
    public void func_150895_a(Item itemIn, CreativeTabs tab, List<ItemStack> subItems)
    {
        for (Fluid fluid : FluidRegistry.getRegisteredFluids().values())
        {
            // add all fluids that the bucket can be filled  with
            FluidStack fs = new FluidStack(fluid, capacity);
            ItemStack stack = new ItemStack(this);
            if (fill(stack, fs, true) == fs.amount)
            {
                subItems.add(stack);
            }
        }
    }

    @Override
    public String func_77653_i(ItemStack stack)
    {
        FluidStack fluidStack = getFluid(stack);
        if (fluidStack == null)
        {
            if(empty != null)
            {
                return empty.func_82833_r();
            }
            return super.func_77653_i(stack);
        }

        String unloc = this.func_77657_g(stack);

        if (I18n.func_94522_b(unloc + "." + fluidStack.getFluid().getName()))
        {
            return I18n.func_74838_a(unloc + "." + fluidStack.getFluid().getName());
        }

        return I18n.func_74837_a(unloc + ".name", fluidStack.getLocalizedName());
    }

    @Override
    public ActionResult<ItemStack> func_77659_a(ItemStack itemstack, World world, EntityPlayer player, EnumHand hand)
    {
        FluidStack fluidStack = getFluid(itemstack);
        // empty bucket shouldn't exist, do nothing since it should be handled by the bucket event
        if (fluidStack == null)
        {
            return ActionResult.newResult(EnumActionResult.PASS, itemstack);
        }

        // clicked on a block?
        RayTraceResult mop = this.func_77621_a(world, player, false);
        if (mop != null && mop.field_72313_a == RayTraceResult.Type.BLOCK)
        {
            BlockPos clickPos = mop.func_178782_a();
            // can we place liquid there?
            if (world.func_175660_a(player, clickPos))
            {
                // the block adjacent to the side we clicked on
                BlockPos targetPos = clickPos.func_177972_a(mop.field_178784_b);

                // can the player place there?
                if (player.func_175151_a(targetPos, mop.field_178784_b, itemstack))
                {
                    // try placing liquid
                    if (this.tryPlaceFluid(fluidStack.getFluid().getBlock(), world, targetPos)
                            && !player.field_71075_bZ.field_75098_d)
                    {
                        // success!
                        player.func_71029_a(StatList.func_188057_b(this));

                        itemstack.field_77994_a--;
                        ItemStack emptyStack = empty != null ? empty.func_77946_l() : new ItemStack(this);

                        // check whether we replace the item or add the empty one to the inventory
                        if (itemstack.field_77994_a <= 0)
                        {
                            return ActionResult.newResult(EnumActionResult.SUCCESS, emptyStack);
                        }
                        else
                        {
                            // add empty bucket to player inventory
                            ItemHandlerHelper.giveItemToPlayer(player, emptyStack);
                            return ActionResult.newResult(EnumActionResult.SUCCESS, itemstack);
                        }
                    }
                }
            }
        }

        // couldn't place liquid there2
        return ActionResult.newResult(EnumActionResult.FAIL, itemstack);
    }


    public boolean tryPlaceFluid(Block block, World worldIn, BlockPos pos)
    {
        if (block == null)
        {
            return false;
        }

        Material material = worldIn.func_180495_p(pos).func_185904_a();
        boolean isSolid = material.func_76220_a();

        // can only place in air or non-solid blocks
        if (!worldIn.func_175623_d(pos) && isSolid)
        {
            return false;
        }

        // water goes poof?
        if (worldIn.field_73011_w.func_177500_n() && (block == Blocks.field_150358_i || block == Blocks.field_150355_j))
        {
            int i = pos.func_177958_n();
            int j = pos.func_177956_o();
            int k = pos.func_177952_p();
            worldIn.func_184133_a(null, pos, SoundEvents.field_187646_bt, SoundCategory.BLOCKS, 0.5F,
                    2.6F + (worldIn.field_73012_v.nextFloat() - worldIn.field_73012_v.nextFloat()) * 0.8F);

            for (int l = 0; l < 8; ++l)
            {
                worldIn.func_175688_a(EnumParticleTypes.SMOKE_LARGE,
                        (double) i + Math.random(),
                        (double) j + Math.random(), (double) k + Math.random(), 0.0D, 0.0D, 0.0D);
            }
        }
        else
        {
            if (!worldIn.field_72995_K && !isSolid && !material.func_76224_d())
            {
                worldIn.func_175655_b(pos, true);
            }

            worldIn.func_180501_a(pos, block.func_176223_P(), 3);
        }
        return true;
    }

    @SubscribeEvent(priority = EventPriority.LOW) // low priority so other mods can handle their stuff first
    public void onFillBucket(FillBucketEvent event)
    {
        if (event.getResult() != Event.Result.DEFAULT)
        {
            // event was already handled
            return;
        }

        // not for us to handle
        if (event.getEmptyBucket() == null ||
                !event.getEmptyBucket().func_77969_a(empty) ||
                (nbtSensitive && ItemStack.func_77970_a(event.getEmptyBucket(), empty)))
        {
            return;
        }

        // needs to target a block
        if (event.getTarget() == null || event.getTarget().field_72313_a != RayTraceResult.Type.BLOCK)
        {
            return;
        }

        World world = event.getWorld();
        BlockPos pos = event.getTarget().func_178782_a();
        IBlockState state = world.func_180495_p(pos);
        // Note that water and lava are NOT an instance of IFluidBlock! They are therefore not handled by this code!
        if (state.func_177230_c() instanceof IFluidBlock)
        {
            IFluidBlock fluidBlock = (IFluidBlock) state.func_177230_c();
            if (fluidBlock.canDrain(world, pos))
            {
                FluidStack drained = fluidBlock.drain(world, pos, false);
                // check if it fits exactly
                if (drained != null && drained.amount == capacity)
                {
                    // check if the container accepts it
                    ItemStack filledBucket = new ItemStack(this);
                    int filled = this.fill(filledBucket, drained, false);
                    if (filled == drained.amount)
                    {
                        // actually transfer the fluid
                        drained = fluidBlock.drain(world, pos, true);
                        this.fill(filledBucket, drained, true);

                        // set it as the result
                        event.setResult(Event.Result.ALLOW);
                        event.setFilledBucket(filledBucket);
                    }
                    else
                    {
                        // cancel event, otherwise the vanilla minecraft ItemBucket would
                        // convert it into a water/lava bucket depending on the blocks material
                        event.setCanceled(true);
                    }
                }
            }
        }
    }

    public static ItemStack getFilledBucket(UniversalBucket item, Fluid fluid)
    {
        ItemStack stack = new ItemStack(item);
        item.fill(stack, new FluidStack(fluid, item.capacity), true);
        return stack;
    }

  /* FluidContainer Management */

    @Override
    public FluidStack getFluid(ItemStack container)
    {
        return FluidStack.loadFluidStackFromNBT(container.func_77978_p());
    }

    @Override
    public int getCapacity(ItemStack container)
    {
        return capacity;
    }

    @Override
    public int fill(ItemStack container, FluidStack resource, boolean doFill)
    {
        // has to be exactly 1, must be handled from the caller
        if (container.field_77994_a != 1)
        {
            return 0;
        }

        // can only fill exact capacity
        if (resource == null || resource.amount != capacity)
        {
            return 0;
        }
        // registered in the registry?
        if (!FluidRegistry.getBucketFluids().contains(resource.getFluid()))
        {
            return 0;
        }
        // fill the container
        if (doFill)
        {
            NBTTagCompound tag = container.func_77978_p();
            if (tag == null)
            {
                tag = new NBTTagCompound();
            }
            resource.writeToNBT(tag);
            container.func_77982_d(tag);
        }
        return capacity;
    }

    @Override
    public FluidStack drain(ItemStack container, int maxDrain, boolean doDrain)
    {
        // can only drain everything at once
        if (maxDrain < capacity)
        {
            return null;
        }

        FluidStack fluidStack = getFluid(container);
        if (doDrain && fluidStack != null)
        {
            if(empty != null)
            {
                container.func_150996_a(empty.func_77973_b());
                container.func_77982_d(empty.func_77978_p());
                container.func_77964_b(empty.func_77952_i());
            }
            else {
                container.field_77994_a = 0;
            }
        }

        return fluidStack;
    }
}
